@@ -91,14 +91,18 @@ struct cpudef {
 /*
   ( addr -- ? )
 */
+
+extern volatile uint64_t client_tba;
+
 static void
 set_trap_table(void)
 {
     unsigned long addr;
 
     addr = POP();
-    asm("wrpr %0, %%tba\n"
-        : : "r" (addr));
+
+    /* Update client_tba to be updated on CIF exit */
+    client_tba = addr;
 }
 
 /* Reset control register is defined in 17.2.7.3 of US IIi User Manual */
@@ -112,6 +116,20 @@ sparc64_reset_all(void)
         : : "r" (val), "r" (addr) : "memory");
 }
 
+/* PCI Target Address Space Register (see UltraSPARC IIi User's Manual
+  section 19.3.0.4) */
+#define PBM_PCI_TARGET_AS              0x2028
+#define PBM_PCI_TARGET_AS_CD_ENABLE    0x40
+
+static void
+sparc64_set_tas_register(unsigned long val)
+{
+    unsigned long addr = APB_SPECIAL_BASE + PBM_PCI_TARGET_AS;
+
+    asm("stxa %0, [%1] 0x15\n\t"
+        : : "r" (val), "r" (addr) : "memory");
+}
+
 static void cpu_generic_init(const struct cpudef *cpu, uint32_t clock_frequency)
 {
     unsigned long iu_version;
@@ -561,6 +579,8 @@ static void init_memory(void)
     PUSH(virt + MEMORY_SIZE);
 }
 
+extern volatile uint64_t *obp_ticks_pointer;
+
 static void
 arch_init( void )
 {
@@ -568,10 +588,18 @@ arch_init( void )
 	modules_init();
 #ifdef CONFIG_DRIVER_PCI
         ob_pci_init();
+
+        /* Set TAS register to match the virtual-dma properties
+           set during sabre configure */
+        sparc64_set_tas_register(PBM_PCI_TARGET_AS_CD_ENABLE);
 #endif
         nvconf_init();
         device_end();
 
+        /* Point to the Forth obp-ticks variable */
+        fword("obp-ticks");
+        obp_ticks_pointer = cell2pointer(POP());
+
 	bind_func("platform-boot", boot );
 	bind_func("(go)", go);
 }
